from __future__ import annotations

from loxygen.token import Token


class LoxError(Exception):
    pass


class LoxParseError(LoxError):
    pass


class LoxRunTimeError(LoxError):
    def __init__(self, token, message):
        super().__init__()
        self.token: Token = token
        self.message: str = message



from __future__ import annotations

from enum import IntEnum
from os import EX_DATAERR
from os import EX_OK
from os import EX_SOFTWARE


class LoxStatus(IntEnum):
    OK = EX_OK
    STATIC_ERROR = EX_DATAERR
    RUNTIME_ERROR = EX_SOFTWARE



from __future__ import annotations

from loxygen.token import Token
from loxygen.token import TokenType

IDENTIFIERS = [
    "AND",
    "CLASS",
    "ELSE",
    "FALSE",
    "FUN",
    "FOR",
    "IF",
    "NIL",
    "OR",
    "PRINT",
    "RETURN",
    "SUPER",
    "THIS",
    "TRUE",
    "VAR",
    "WHILE",
]


class Scanner:
    def __init__(self, source: str):
        self.source = source
        self.tokens: list[Token] = []

        self.start = 0
        self.current = 0
        self.line = 1

        self.keywords = {
            token_type.name.lower(): token_type
            for token_type in TokenType
            if token_type.name in IDENTIFIERS
        }

        self.errors: list[tuple[int, str]] = []

    def scan_tokens(self):
        while not self.is_at_end():
            self.start = self.current
            self.scan_token()

        self.tokens.append(Token(TokenType.EOF, "", None, self.line))

    def scan_token(self):
        char = self.advance()
        match char:
            case "(":
                self.add_token(TokenType.LEFT_PAREN)
            case ")":
                self.add_token(TokenType.RIGHT_PAREN)
            case "{":
                self.add_token(TokenType.LEFT_BRACE)
            case "}":
                self.add_token(TokenType.RIGHT_BRACE)
            case ",":
                self.add_token(TokenType.COMMA)
            case ".":
                self.add_token(TokenType.DOT)
            case "-":
                self.add_token(TokenType.MINUS)
            case "+":
                self.add_token(TokenType.PLUS)
            case ";":
                self.add_token(TokenType.SEMICOLON)
            case "*":
                self.add_token(TokenType.STAR)
            case "!":
                self.add_token(
                    TokenType.BANG_EQUAL if self.match("=") else TokenType.BANG,
                )
            case "=":
                self.add_token(
                    TokenType.EQUAL_EQUAL if self.match("=") else TokenType.EQUAL,
                )
            case "<":
                self.add_token(
                    TokenType.LESS_EQUAL if self.match("=") else TokenType.LESS,
                )
            case ">":
                self.add_token(
                    TokenType.GREATER_EQUAL if self.match("=") else TokenType.GREATER,
                )
            case "/":
                if self.match("/"):
                    while self.peek() != "\n" and not self.is_at_end():
                        self.advance()
                else:
                    self.add_token(TokenType.SLASH)
            case " " | "\r" | "\t":
                pass
            case "\n":
                self.line += 1
            case '"':
                self.string()
            case _:
                if self.is_digit(char):
                    self.number()
                elif self.is_alpha(char):
                    self.identifier()
                else:
                    self.errors.append((self.line, "Error: Unexpected character."))

    def identifier(self):
        while self.is_alphanumeric(self.peek()):
            self.advance()
        text = self.source[self.start : self.current]
        token_type = self.keywords.get(text)
        if token_type is None:
            token_type = TokenType.IDENTIFIER
        self.add_token(token_type)

    def number(self):
        while self.is_digit(self.peek()):
            self.advance()
        if self.peek() == "." and self.is_digit(self.peek_next()):
            self.advance()
            while self.is_digit(self.peek()):
                self.advance()
        self.add_token(TokenType.NUMBER, float(self.source[self.start : self.current]))

    def string(self):
        while self.peek() != '"' and not self.is_at_end():
            if self.peek() == "\n":
                self.line += 1
            self.advance()

        if self.is_at_end():
            self.errors.append((self.line, "Error: Unterminated string."))
            return

        self.advance()
        value = self.source[self.start + 1 : self.current - 1]
        self.add_token(TokenType.STRING, value)

    def match(self, expected: str):
        if self.is_at_end() or self.source[self.current] != expected:
            return False
        self.current += 1
        return True

    def peek(self):
        if self.is_at_end():
            return ""
        return self.source[self.current]

    def peek_next(self):
        if self.current + 1 >= len(self.source):
            return ""
        return self.source[self.current + 1]

    @staticmethod
    def is_alpha(char: str):
        return (char >= "a" and char <= "z") or (char >= "A" and char <= "Z") or char == "_"

    @staticmethod
    def is_digit(char: str):
        return char >= "0" and char <= "9"

    def is_alphanumeric(self, char: str):
        return self.is_alpha(char) or self.is_digit(char)

    def is_at_end(self):
        return self.current >= len(self.source)

    def advance(self):
        char = self.source[self.current]
        self.current += 1
        return char

    def add_token(self, type: TokenType, literal=None):
        text = self.source[self.start : self.current]
        self.tokens.append(Token(type, text, literal, self.line))



from __future__ import annotations

from loxygen import nodes
from loxygen.exceptions import LoxParseError
from loxygen.token import Token
from loxygen.token import TokenType

MAXIMUM_ARGS_NUMBER = 255


class Parser:
    def __init__(self, tokens: list[Token]):
        self.tokens = tokens
        self.current = 0
        self.errors: list[tuple[Token, str]] = []

    def parse(self):
        statements = []
        while not self.is_at_end():
            statements.append(self.declaration())

        return statements

    def expression(self) -> nodes.Expr:
        return self.assignment()

    def declaration(self):
        try:
            if self.match(TokenType.CLASS):
                return self.class_declaration()
            if self.match(TokenType.FUN):
                return self.function("function")
            if self.match(TokenType.VAR):
                return self.var_declaration()
            return self.statement()
        except LoxParseError:
            self.synchronize()
            return None

    def class_declaration(self):
        name = self.consume(TokenType.IDENTIFIER, "Expect class name.")

        superclass = None
        if self.match(TokenType.LESS):
            self.consume(TokenType.IDENTIFIER, "Expect superclass name.")
            superclass = nodes.Variable(self.previous())

        self.consume(TokenType.LEFT_BRACE, "Expect '{' before class body.")

        methods = []
        while not (self.check(TokenType.RIGHT_BRACE) or self.is_at_end()):
            methods.append(self.function("method"))

        self.consume(TokenType.RIGHT_BRACE, "Expect '}' after class body.")

        return nodes.Class(name, superclass, methods)

    def statement(self):
        if self.match(TokenType.FOR):
            return self.for_statement()
        if self.match(TokenType.IF):
            return self.if_statement()
        if self.match(TokenType.PRINT):
            return self.print_statement()
        if self.match(TokenType.RETURN):
            return self.return_statement()
        if self.match(TokenType.WHILE):
            return self.while_statement()
        if self.match(TokenType.LEFT_BRACE):
            return nodes.Block(self.block())

        return self.expression_statement()

    def for_statement(self):
        self.consume(TokenType.LEFT_PAREN, "Expect '(' after if.")

        if self.match(TokenType.SEMICOLON):
            initializer = None
        elif self.match(TokenType.VAR):
            initializer = self.var_declaration()
        else:
            initializer = self.expression_statement()

        condition = None if self.check(TokenType.SEMICOLON) else self.expression()
        self.consume(TokenType.SEMICOLON, "Expect ';' after loop condition.")

        increment = None if self.check(TokenType.RIGHT_PAREN) else self.expression()
        self.consume(TokenType.RIGHT_PAREN, "Expect ')' after for clauses.")

        body = self.statement()

        if increment is not None:
            body = nodes.Block([body, increment])

        if condition is None:
            condition = nodes.Literal(True)
        body = nodes.While(condition, body)

        if initializer is not None:
            body = nodes.Block([initializer, body])

        return body

    def if_statement(self):
        self.consume(TokenType.LEFT_PAREN, "Expect '(' after if.")
        condition = self.expression()
        self.consume(TokenType.RIGHT_PAREN, "Expect ')' after if condition.")

        then_branch = self.statement()
        else_branch = self.statement() if self.match(TokenType.ELSE) else None

        return nodes.If(condition, then_branch, else_branch)

    def print_statement(self):
        value = self.expression()
        self.consume(TokenType.SEMICOLON, "Expect ';' after value.")

        return nodes.Print(value)

    def return_statement(self):
        keyword = self.previous()
        value = None if self.check(TokenType.SEMICOLON) else self.expression()
        self.consume(TokenType.SEMICOLON, "Expect ';' after return value.")

        return nodes.Return(keyword, value)

    def var_declaration(self):
        name = self.consume(TokenType.IDENTIFIER, "Expect variable name.")

        initializer = None
        if self.match(TokenType.EQUAL):
            initializer = self.expression()

        self.consume(TokenType.SEMICOLON, "Expect ';' after expression.")
        return nodes.Var(name, initializer)

    def while_statement(self):
        self.consume(TokenType.LEFT_PAREN, "Expect '(' after if.")
        condition = self.expression()
        self.consume(TokenType.RIGHT_PAREN, "Expect ')' after if condition.")

        body = self.statement()

        return nodes.While(condition, body)

    def expression_statement(self):
        expression = self.expression()
        self.consume(TokenType.SEMICOLON, "Expect ';' after expression.")

        return nodes.Expression(expression)

    def function(self, kind: str) -> nodes.Function:
        name = self.consume(TokenType.IDENTIFIER, f"Expect {kind} name.")
        self.consume(TokenType.LEFT_PAREN, f"Expect '(' after {kind} name.")
        parameters: list[Token] = []
        if not self.check(TokenType.RIGHT_PAREN):
            while True:
                if len(parameters) >= MAXIMUM_ARGS_NUMBER:
                    self.error(
                        self.peek(),
                        f"Can't have more than {MAXIMUM_ARGS_NUMBER} parameters.",
                    )

                parameters.append(
                    self.consume(TokenType.IDENTIFIER, "Expect parameter name"),
                )
                if not self.match(TokenType.COMMA):
                    break

        self.consume(TokenType.RIGHT_PAREN, "Expect ')' after parameters.")
        self.consume(TokenType.LEFT_BRACE, f"Expect '{{' before {kind} body.")
        body = self.block()

        return nodes.Function(name, parameters, body)

    def block(self):
        statements = []
        while not (self.check(TokenType.RIGHT_BRACE) or self.is_at_end()):
            statements.append(self.declaration())

        self.consume(TokenType.RIGHT_BRACE, "Expect '}' after expression.")
        return statements

    def assignment(self) -> nodes.Expr:
        expression = self.logical_or()
        if self.match(TokenType.EQUAL):
            equals = self.previous()
            value = self.assignment()
            if isinstance(expression, nodes.Variable):
                return nodes.Assign(expression.name, value)
            elif isinstance(expression, nodes.Get):
                return nodes.Set(expression.object, expression.name, value)
            self.error(equals, "Invalid assignment target.")
        return expression

    def logical_or(self) -> nodes.Expr:
        expression = self.logical_and()
        while self.match(TokenType.OR):
            operator = self.previous()
            right = self.logical_and()
            expression = nodes.Logical(expression, operator, right)

        return expression

    def logical_and(self) -> nodes.Expr:
        expression = self.equality()
        while self.match(TokenType.AND):
            operator = self.previous()
            right = self.equality()
            expression = nodes.Logical(expression, operator, right)

        return expression

    def equality(self) -> nodes.Expr:
        expression = self.comparison()
        while self.match(TokenType.BANG_EQUAL, TokenType.EQUAL_EQUAL):
            operator = self.previous()
            right = self.comparison()
            expression = nodes.Binary(expression, operator, right)

        return expression

    def comparison(self) -> nodes.Expr:
        expression = self.term()
        while self.match(
            TokenType.GREATER,
            TokenType.GREATER_EQUAL,
            TokenType.LESS,
            TokenType.LESS_EQUAL,
        ):
            operator = self.previous()
            right = self.term()
            expression = nodes.Binary(expression, operator, right)

        return expression

    def term(self) -> nodes.Expr:
        expression = self.factor()
        while self.match(TokenType.MINUS, TokenType.PLUS):
            operator = self.previous()
            right = self.factor()
            expression = nodes.Binary(expression, operator, right)

        return expression

    def factor(self) -> nodes.Expr:
        expression = self.unary()
        while self.match(TokenType.SLASH, TokenType.STAR):
            operator = self.previous()
            right = self.unary()
            expression = nodes.Binary(expression, operator, right)

        return expression

    def unary(self) -> nodes.Expr:
        if self.match(TokenType.BANG, TokenType.MINUS):
            operator = self.previous()
            right = self.unary()
            return nodes.Unary(operator, right)

        return self.call()

    def finish_call(self, callee) -> nodes.Expr:
        arguments: list[nodes.Expr] = []
        if not self.check(TokenType.RIGHT_PAREN):
            while True:
                if len(arguments) >= MAXIMUM_ARGS_NUMBER:
                    self.error(
                        self.peek(),
                        f"Can't have more than {MAXIMUM_ARGS_NUMBER} arguments.",
                    )
                arguments.append(self.expression())
                if not self.match(TokenType.COMMA):
                    break

        paren = self.consume(TokenType.RIGHT_PAREN, "Expect ')' after arguments.")

        return nodes.Call(callee, paren, arguments)

    def call(self) -> nodes.Expr:
        expression = self.primary()
        while True:
            if self.match(TokenType.LEFT_PAREN):
                expression = self.finish_call(expression)
            elif self.match(TokenType.DOT):
                name = self.consume(
                    TokenType.IDENTIFIER,
                    "Expect property name after '.'.",
                )
                expression = nodes.Get(expression, name)
            else:
                break

        return expression

    def primary(self) -> nodes.Expr:
        if self.match(TokenType.TRUE):
            return nodes.Literal(True)
        if self.match(TokenType.FALSE):
            return nodes.Literal(False)
        if self.match(TokenType.NIL):
            return nodes.Literal(None)
        if self.match(TokenType.NUMBER, TokenType.STRING):
            return nodes.Literal(self.previous().literal)
        if self.match(TokenType.THIS):
            return nodes.This(self.previous())
        if self.match(TokenType.IDENTIFIER):
            return nodes.Variable(self.previous())
        if self.match(TokenType.SUPER):
            keyword = self.previous()
            self.consume(TokenType.DOT, "Expect '.' after 'super'.")
            method = self.consume(TokenType.IDENTIFIER, "Expect superclass method name.")
            return nodes.Super(keyword, method)
        if self.match(TokenType.LEFT_PAREN):
            expression = self.expression()
            self.consume(TokenType.RIGHT_PAREN, "Expect ')' after expression.")
            return nodes.Grouping(expression)

        raise self.error(self.peek(), "Expect expression.")

    def match(self, *types: TokenType) -> bool:
        if check := self.check(*types):
            self.advance()
        return check

    def consume(self, token_type: TokenType, message: str) -> Token:
        if self.check(token_type):
            return self.advance()
        raise self.error(self.peek(), message)

    def check(self, *types: TokenType) -> bool:
        if self.is_at_end():
            return False
        return self.peek().type in types

    def advance(self) -> Token:
        if not self.is_at_end():
            self.current += 1
        return self.previous()

    def is_at_end(self) -> bool:
        return self.peek().type == TokenType.EOF

    def peek(self) -> Token:
        return self.tokens[self.current]

    def previous(self) -> Token:
        return self.tokens[self.current - 1]

    def error(self, token: Token, message: str) -> LoxParseError:
        self.errors.append((token, message))
        return LoxParseError(message, token)

    def synchronize(self):
        self.advance()
        while not self.is_at_end():
            if self.previous().type == TokenType.SEMICOLON or self.check(
                TokenType.CLASS,
                TokenType.FUN,
                TokenType.VAR,
                TokenType.FOR,
                TokenType.IF,
                TokenType.WHILE,
                TokenType.PRINT,
                TokenType.RETURN,
            ):
                return
            self.advance()



from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass

from loxygen.token import Token


class Visitor(ABC):
    @abstractmethod
    def visit_assign_expr(self, expr: Assign):
        pass

    @abstractmethod
    def visit_binary_expr(self, expr: Binary):
        pass

    @abstractmethod
    def visit_call_expr(self, expr: Call):
        pass

    @abstractmethod
    def visit_get_expr(self, expr: Get):
        pass

    @abstractmethod
    def visit_grouping_expr(self, expr: Grouping):
        pass

    @abstractmethod
    def visit_literal_expr(self, expr: Literal):
        pass

    @abstractmethod
    def visit_logical_expr(self, expr: Logical):
        pass

    @abstractmethod
    def visit_set_expr(self, expr: Set):
        pass

    @abstractmethod
    def visit_super_expr(self, expr: Super):
        pass

    @abstractmethod
    def visit_this_expr(self, expr: This):
        pass

    @abstractmethod
    def visit_unary_expr(self, expr: Unary):
        pass

    @abstractmethod
    def visit_variable_expr(self, expr: Variable):
        pass

    @abstractmethod
    def visit_block_stmt(self, stmt: Block):
        pass

    @abstractmethod
    def visit_expression_stmt(self, stmt: Expression):
        pass

    @abstractmethod
    def visit_function_stmt(self, stmt: Function):
        pass

    @abstractmethod
    def visit_class_stmt(self, stmt: Class):
        pass

    @abstractmethod
    def visit_if_stmt(self, stmt: If):
        pass

    @abstractmethod
    def visit_print_stmt(self, stmt: Print):
        pass

    @abstractmethod
    def visit_return_stmt(self, stmt: Return):
        pass

    @abstractmethod
    def visit_var_stmt(self, stmt: Var):
        pass

    @abstractmethod
    def visit_while_stmt(self, stmt: While):
        pass


@dataclass(frozen=True, slots=True)
class Expr:
    def accept(self, visitor: Visitor):
        pass


@dataclass(frozen=True, slots=True)
class Stmt:
    def accept(self, visitor: Visitor):
        pass


@dataclass(frozen=True, slots=True)
class Assign(Expr):
    name: Token
    value: Expr

    def accept(self, visitor: Visitor):
        return visitor.visit_assign_expr(self)


@dataclass(frozen=True, slots=True)
class Binary(Expr):
    left: Expr
    operator: Token
    right: Expr

    def accept(self, visitor: Visitor):
        return visitor.visit_binary_expr(self)


@dataclass(frozen=True, slots=True)
class Call(Expr):
    callee: Expr
    paren: Token
    arguments: list[Expr]

    def accept(self, visitor: Visitor):
        return visitor.visit_call_expr(self)


@dataclass(frozen=True, slots=True)
class Get(Expr):
    object: Expr
    name: Token

    def accept(self, visitor: Visitor):
        return visitor.visit_get_expr(self)


@dataclass(frozen=True, slots=True)
class Grouping(Expr):
    expression: Expr

    def accept(self, visitor: Visitor):
        return visitor.visit_grouping_expr(self)


@dataclass(frozen=True, slots=True)
class Literal(Expr):
    value: object

    def accept(self, visitor: Visitor):
        return visitor.visit_literal_expr(self)


@dataclass(frozen=True, slots=True)
class Logical(Expr):
    left: Expr
    operator: Token
    right: Expr

    def accept(self, visitor: Visitor):
        return visitor.visit_logical_expr(self)


@dataclass(frozen=True, slots=True)
class Set(Expr):
    object: Expr
    name: Token
    value: Expr

    def accept(self, visitor: Visitor):
        return visitor.visit_set_expr(self)


@dataclass(frozen=True, slots=True)
class Super(Expr):
    keyword: Token
    method: Token

    def accept(self, visitor: Visitor):
        return visitor.visit_super_expr(self)


@dataclass(frozen=True, slots=True)
class This(Expr):
    keyword: Token

    def accept(self, visitor: Visitor):
        return visitor.visit_this_expr(self)


@dataclass(frozen=True, slots=True)
class Unary(Expr):
    operator: Token
    right: Expr

    def accept(self, visitor: Visitor):
        return visitor.visit_unary_expr(self)


@dataclass(frozen=True, slots=True)
class Variable(Expr):
    name: Token

    def accept(self, visitor: Visitor):
        return visitor.visit_variable_expr(self)


@dataclass(frozen=True, slots=True)
class Block(Stmt):
    statements: list[Stmt]

    def accept(self, visitor: Visitor):
        return visitor.visit_block_stmt(self)


@dataclass(frozen=True, slots=True)
class Expression(Stmt):
    expression: Expr

    def accept(self, visitor: Visitor):
        return visitor.visit_expression_stmt(self)


@dataclass(frozen=True, slots=True)
class Function(Stmt):
    name: Token
    params: list[Token]
    body: list[Stmt]

    def accept(self, visitor: Visitor):
        return visitor.visit_function_stmt(self)


@dataclass(frozen=True, slots=True)
class Class(Stmt):
    name: Token
    superclass: Variable
    methods: list[Function]

    def accept(self, visitor: Visitor):
        return visitor.visit_class_stmt(self)


@dataclass(frozen=True, slots=True)
class If(Stmt):
    condition: Expr
    then_branch: Stmt
    else_branch: Stmt

    def accept(self, visitor: Visitor):
        return visitor.visit_if_stmt(self)


@dataclass(frozen=True, slots=True)
class Print(Stmt):
    expression: Expr

    def accept(self, visitor: Visitor):
        return visitor.visit_print_stmt(self)


@dataclass(frozen=True, slots=True)
class Return(Stmt):
    keyword: Token
    value: Expr

    def accept(self, visitor: Visitor):
        return visitor.visit_return_stmt(self)


@dataclass(frozen=True, slots=True)
class Var(Stmt):
    name: Token
    initializer: Expr

    def accept(self, visitor: Visitor):
        return visitor.visit_var_stmt(self)


@dataclass(frozen=True, slots=True)
class While(Stmt):
    condition: Expr
    body: Stmt

    def accept(self, visitor: Visitor):
        return visitor.visit_while_stmt(self)



from __future__ import annotations

import readline  # noqa: F401
import sys
from os import EX_USAGE

from contract.contract import LoxStatus
from loxygen.exceptions import LoxRunTimeError
from loxygen.interpreter import Interpreter
from loxygen.parser import Parser
from loxygen.resolver import Resolver
from loxygen.scanner import Scanner
from loxygen.token import Token
from loxygen.token import TokenType


class Lox:
    def __init__(self):
        self.interpreter = Interpreter()

    def main(self, args: list[str]):
        if len(args) > 1:
            print("Usage: loxygen [script]")
            sys.exit(EX_USAGE)
        elif len(args) == 1:
            self.run_file(args[0])
        else:
            self.run_prompt()

    def run_file(self, filename: str):
        with open(filename) as f:
            source = f.read()
        status = self.run(source)
        if status != LoxStatus.OK:
            sys.exit(status.value)

    def run_prompt(self):
        while True:
            line = input(">")
            if line == "":
                break
            self.run(line)

    def run(self, source: str):
        scanner = Scanner(source)
        scanner.scan_tokens()
        parser = Parser(scanner.tokens)
        statements = parser.parse()

        if self.check_error(scanner.errors + parser.errors):
            return LoxStatus.STATIC_ERROR

        resolver = Resolver(self.interpreter)
        resolver.resolve(*statements)

        if self.check_error(resolver.errors):
            return LoxStatus.STATIC_ERROR

        try:
            self.interpreter.interpret(*statements)
        except LoxRunTimeError as e:
            self.error(e.token.line, e.message)
            return LoxStatus.RUNTIME_ERROR

        return LoxStatus.OK

    def check_error(self, errors):
        error = False
        for context, message in errors:
            self.error(context, message)
            error = True
        return error

    def error(
        self,
        context: int | Token,
        message: str,
    ):
        if isinstance(context, int):
            line = context
            token = ""
        elif isinstance(context, Token):
            line = context.line
            position = "end" if context.type == TokenType.EOF else f"'{context.lexeme}'"
            token = f"Error at {position}: "

        print(f"[line {line}] {token}{message}", file=sys.stderr)


def main():
    Lox().main(sys.argv[1:])


if __name__ == "__main__":
    main()



from __future__ import annotations

import argparse
import re
import shutil
import subprocess
import sys
from pathlib import Path
from signal import SIGINT
from subprocess import CalledProcessError
from tempfile import TemporaryDirectory

URL = "https://github.com/munificent/craftinginterpreters.git"
REMOTE_TESTS_DIR = "test"

LINE_NB_RULES: dict[str, dict[str, str]] = {
    "static": {r"(//)( Error)": r"\1 [line {line_number}]\2"},
    "runtime": {r"(// expect runtime error:)(?! \[line \d+])": r"\1 [line {line_number}]"},
}

PREFIX_RULE: dict[str, str] = {r"\[{prefix} (line \d+)]": r"[\1]"}


class CommandError(Exception):
    def __init__(self, message: str, exit_code: int = 1):
        super().__init__(message)
        self.exit_code = exit_code


def download(destination: Path, url: str, remote_tests_dir: str, force: bool = False) -> None:
    if destination.exists() and not force:
        raise FileExistsError(f"Destination '{destination}' already exists.")

    if (git := shutil.which("git")) is None:
        raise FileNotFoundError("Command 'git' not found.")

    print(f"Downloading tests to '{destination}'...")
    with TemporaryDirectory() as temp_dir:
        commands = [
            [git, "clone", "--no-checkout", "--depth=1", "--filter=blob:none", url, "."],
            [git, "sparse-checkout", "set", "--no-cone", remote_tests_dir],
            [git, "checkout"],
        ]
        for cmd in commands:
            subprocess.run(
                cmd,
                cwd=temp_dir,
                check=True,
                capture_output=True,
                text=True,
            )

        if destination.exists():
            shutil.rmtree(destination)

        source = Path(temp_dir) / remote_tests_dir
        shutil.move(source, destination)


def process_rules(
    line_nb_rules: dict[str, dict[str, str]],
    prefix_rule: dict[str, str],
    line_number: list[str],
    prefixes: list[str],
) -> dict[re.Pattern[str], str]:
    result: dict[str, str] = {}

    result.update(
        {
            pattern: replacement
            for mode in line_number
            for pattern, replacement in line_nb_rules[mode].items()
        }
    )
    result.update(
        {
            pattern.format(prefix=prefix): replacement
            for prefix in prefixes
            for pattern, replacement in prefix_rule.items()
        }
    )

    return {re.compile(pattern): replacement for pattern, replacement in result.items()}


def process_file(path: Path, rules: dict[re.Pattern[str], str]) -> None:
    text = path.read_text()
    has_trailing_newline = text.endswith("\n")
    lines = text.splitlines()
    for nb, line in enumerate(lines, 1):
        for pattern, replacement in rules.items():
            line = re.sub(pattern, replacement.format(line_number=nb), line)
        lines[nb - 1] = line

    text = "\n".join(lines) + "\n" * has_trailing_newline
    path.write_text(text)


def process_directory(
    path: Path,
    line_nb_rules: dict[str, dict[str, str]],
    prefix_rule: dict[str, str],
    line_number: list[str],
    prefixes: list[str],
) -> None:
    if not path.exists():
        raise FileNotFoundError(f"Directory '{path}' does not exist.")
    if not path.is_dir():
        raise ValueError(f"Path '{path}' is not a directory.")
    if not (paths := list(path.rglob("*.lox"))):
        raise FileNotFoundError(f"No '.lox' files found in '{path}'.")

    print(f"Processing files in '{path}'...")
    active_rules = process_rules(line_nb_rules, prefix_rule, line_number, prefixes)
    for path in paths:
        process_file(path, active_rules)
    print("Processing complete.")


def handle_download(args: argparse.Namespace) -> None:
    download(
        args.directory,
        args.url,
        args.remote_tests_dir,
        args.force,
    )


def handle_process(args: argparse.Namespace) -> None:
    process_directory(
        args.directory,
        args.line_nb_rules,
        args.prefix_rule,
        args.line_number,
        args.prefix,
    )


def handle_run(args: argparse.Namespace, *unknown_args: str) -> None:
    cmd = ["pytest"]
    if args.pytest_help:
        cmd.append("--help")
    else:
        if args.interpreter_cmd:
            cmd.extend(["--interpreter_cmd", args.interpreter_cmd])
        if args.skip_dirs:
            cmd.extend([item for dir in args.skip_dirs for item in ("--skip_dirs", dir)])

    cmd.extend(list(unknown_args))

    try:
        process = subprocess.run(cmd)
    except OSError as e:
        # exit code that doesn't conflict with pytest's exit codes (0-5)
        raise CommandError(str(e), exit_code=6) from e

    sys.exit(process.returncode)


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=("A toolkit for downloading and running the official Lox test suite."),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparsers = parser.add_subparsers(dest="command", required=True)

    path_parser = argparse.ArgumentParser(add_help=False)
    path_parser.add_argument(
        "directory",
        type=Path,
        metavar="PATH",
        help="Directory where the operation will be performed.",
    )

    parser_download = subparsers.add_parser(
        "download",
        parents=[path_parser],
        help="Download the official Lox test suite.",
    )
    parser_download.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Overwrite destination directory if it exists.",
    )
    parser_download.set_defaults(
        func=handle_download,
        url=URL,
        remote_tests_dir=REMOTE_TESTS_DIR,
    )

    parser_process = subparsers.add_parser(
        "process",
        parents=[path_parser],
        help="Process an existing directory of Lox tests.",
    )
    parser_process.add_argument(
        "-l",
        "--line_number",
        action="append",
        metavar="MODE",
        choices=["static", "runtime"],
        default=[],
        help="Add '[line N]' to errors. (Choices: 'static', 'runtime').",
    )
    parser_process.add_argument(
        "-p",
        "--prefix",
        action="append",
        metavar="PREFIX",
        choices=["java", "c"],
        default=[],
        help="Remove a language prefix. (Choices: 'java', 'c').",
    )
    parser_process.set_defaults(
        func=handle_process, line_nb_rules=LINE_NB_RULES, prefix_rule=PREFIX_RULE
    )

    parser_run = subparsers.add_parser(
        "run",
        help="Run tests against a Lox interpreter.",
    )
    parser_run.add_argument(
        "--pytest-help",
        action="store_true",
        help="Display pytest's help message and exit.",
    )
    parser_run.add_argument(
        "-i",
        "--interpreter_cmd",
        metavar="CMD",
        help="The command to run the interpreter.",
    )
    parser_run.add_argument(
        "-s",
        "--skip_dirs",
        action="append",
        metavar="DIR",
        help="Skip tests within the specified subdirectory (e.g., benchmark)."
        " Can be specified multiple times.",
    )
    parser_run.set_defaults(func=handle_run)

    return parser


def get_parsed_arguments() -> tuple[argparse.Namespace, list[str]]:
    parser = get_parser()
    args, unknown_args = parser.parse_known_args()
    if args.command != "run" and unknown_args:
        parser.parse_args()
    if args.command == "process" and not (args.line_number or args.prefix):
        parser.error("no transformation specified.")

    return args, unknown_args


def main() -> None:
    args, unknown_args = get_parsed_arguments()

    try:
        args.func(args, *unknown_args)
    except (CalledProcessError, CommandError, OSError, ValueError) as e:
        exit_code = e.exit_code if isinstance(e, CommandError) else 1

        msg = f"Error: {e}"
        if isinstance(e, CalledProcessError):
            msg = f"{e.stderr.strip()}\n{msg}"

        print(msg, file=sys.stderr)
        sys.exit(exit_code)
    except KeyboardInterrupt:
        print("\nOperation aborted by user.", file=sys.stderr)
        sys.exit(128 + SIGINT)


if __name__ == "__main__":
    main()



from __future__ import annotations

from time import perf_counter_ns

from loxygen import nodes
from loxygen.environment import Environment
from loxygen.exceptions import LoxRunTimeError
from loxygen.token import Token


class Return(RuntimeError):
    def __init__(self, value):
        super().__init__()
        self.value = value


class Callable:
    def call(self, interpreter, arguments):
        pass

    def arity(self):
        pass


class LoxFunction(Callable):
    def __init__(
        self,
        declaration: nodes.Function,
        closure: Environment,
        is_initializer: bool,
    ):
        self.is_initializer = is_initializer
        self.declaration = declaration
        self.closure = closure

    def bind(self, instance):
        env = Environment(self.closure)
        env.define("this", instance)

        return LoxFunction(self.declaration, env, self.is_initializer)

    def call(self, interpreter, arguments):
        env = Environment(self.closure)
        for param, arg in zip(self.declaration.params, arguments, strict=False):
            env.define(param.lexeme, arg)
        try:
            interpreter.execute_block(self.declaration.body, env)
        except Return as e:
            if self.is_initializer:
                return self.closure.get_at(0, "this")
            return e.value

        if self.is_initializer:
            return self.closure.get_at(0, "this")

    def arity(self):
        return len(self.declaration.params)

    def __repr__(self):
        return f"<fn {self.declaration.name.lexeme}>"


class LoxClass(Callable):
    def __init__(self, name, superclass, methods):
        self.name = name
        self.superclass = superclass
        self.methods = methods

    def find_method(self, name: str):
        if name in self.methods:
            return self.methods[name]

        if self.superclass is not None:
            return self.superclass.find_method(name)

        return None

    def call(self, interpreter, arguments):
        instance = LoxInstance(self)
        if (initializer := self.find_method("init")) is not None:
            initializer.bind(instance).call(interpreter, arguments)
        return instance

    def arity(self):
        initializer = self.find_method("init")
        if initializer is None:
            return 0
        return initializer.arity()

    def __repr__(self):
        return self.name


class LoxInstance:
    def __init__(self, cls):
        self.cls = cls
        self.fields = {}

    def get(self, name: Token):
        if (field := self.fields.get(name.lexeme)) is not None:
            return field

        method = self.cls.find_method(name.lexeme)
        if method is not None:
            return method.bind(self)

        raise LoxRunTimeError(name, f"Undefined property '{name.lexeme}'.")

    def set(self, name: Token, value):
        self.fields[name.lexeme] = value

    def __repr__(self):
        return f"{self.cls.name} instance"


class Clock(Callable):
    @staticmethod
    def call(self, interpreter, arguments):
        return perf_counter_ns() * 1000

    def arity(self):
        return 0

    def __repr__(self):
        return "<native fn>"



from __future__ import annotations

from contextlib import contextmanager
from enum import Enum
from enum import auto

from loxygen import nodes
from loxygen.interpreter import Interpreter
from loxygen.token import Token


class FunctionType(Enum):
    NONE = auto()
    FUNCTION = auto()
    INITIALIZER = auto()
    METHOD = auto()


class ClassType(Enum):
    NONE = auto()
    CLASS = auto()
    SUBCLASS = auto()


class Resolver(nodes.Visitor):
    def __init__(self, interpreter: Interpreter):
        self.interpreter = interpreter
        self.scopes: list[dict] = []
        self.current_function = FunctionType.NONE
        self.current_class = ClassType.NONE
        self.errors: list[tuple[Token, str]] = []

    @property
    def current_scope(self):
        return self.scopes[-1]

    def resolve(self, *statements: nodes.Expr | nodes.Stmt):
        for statement in statements:
            statement.accept(self)

    @contextmanager
    def scope(self):
        try:
            self.scopes.append({})
            yield
        finally:
            self.scopes.pop()

    def declare(self, name: Token):
        if self.scopes:
            if name.lexeme in self.current_scope:
                self.errors.append(
                    (
                        name,
                        "Already a variable with this name in this scope.",
                    ),
                )
            self.current_scope[name.lexeme] = False

    def define(self, name: Token):
        if self.scopes:
            self.current_scope[name.lexeme] = True

    def resolve_local(self, expression: nodes.Expr, name: Token):
        for idx, scope in enumerate(reversed(self.scopes)):
            if name.lexeme in scope:
                self.interpreter.resolve(expression, idx)
                break

    def resolve_function(self, function: nodes.Function, function_type: FunctionType):
        enclosing_function = self.current_function
        self.current_function = function_type
        with self.scope():
            for param in function.params:
                self.declare(param)
                self.define(param)
            self.resolve(*function.body)

        self.current_function = enclosing_function

    def resolve_class_body(self, stmt: nodes.Class):
        with self.scope():
            self.current_scope["this"] = True
            for method in stmt.methods:
                declaration = FunctionType.METHOD
                if method.name.lexeme == "init":
                    declaration = FunctionType.INITIALIZER
                self.resolve_function(method, declaration)

    def visit_block_stmt(self, stmt: nodes.Block):
        with self.scope():
            self.resolve(*stmt.statements)

    def visit_class_stmt(self, stmt: nodes.Class):
        enclosing_class = self.current_class
        self.current_class = ClassType.CLASS

        self.declare(stmt.name)
        self.define(stmt.name)

        if stmt.superclass is not None:
            if stmt.name.lexeme == stmt.superclass.name.lexeme:
                self.errors.append(
                    (
                        stmt.superclass.name,
                        "A class can't inherit from itself.",
                    ),
                )

            self.current_class = ClassType.SUBCLASS
            self.resolve(stmt.superclass)
            with self.scope():
                self.current_scope["super"] = True
                self.resolve_class_body(stmt)

        else:
            self.resolve_class_body(stmt)

        self.current_class = enclosing_class

    def visit_expression_stmt(self, stmt: nodes.Expression):
        self.resolve(stmt.expression)

    def visit_function_stmt(self, stmt: nodes.Function):
        self.declare(stmt.name)
        self.define(stmt.name)

        self.resolve_function(stmt, FunctionType.FUNCTION)

    def visit_if_stmt(self, stmt: nodes.If):
        self.resolve(stmt.condition)
        self.resolve(stmt.then_branch)
        if stmt.else_branch:
            self.resolve(stmt.else_branch)

    def visit_print_stmt(self, stmt: nodes.Print):
        self.resolve(stmt.expression)

    def visit_return_stmt(self, stmt: nodes.Return):
        if self.current_function == FunctionType.NONE:
            self.errors.append(
                (
                    stmt.keyword,
                    "Can't return from top-level code.",
                ),
            )
        if stmt.value is not None:
            if self.current_function == FunctionType.INITIALIZER:
                self.errors.append(
                    (
                        stmt.keyword,
                        "Can't return a value from an initializer.",
                    ),
                )
            self.resolve(stmt.value)

    def visit_var_stmt(self, stmt: nodes.Var):
        self.declare(stmt.name)
        if stmt.initializer is not None:
            self.resolve(stmt.initializer)
        self.define(stmt.name)

    def visit_while_stmt(self, stmt: nodes.While):
        self.resolve(stmt.condition)
        self.resolve(stmt.body)

    def visit_assign_expr(self, expression: nodes.Assign):
        self.resolve(expression.value)
        self.resolve_local(expression, expression.name)

    def visit_binary_expr(self, expression: nodes.Binary):
        self.resolve(expression.left)
        self.resolve(expression.right)

    def visit_call_expr(self, expression: nodes.Call):
        self.resolve(expression.callee)
        for argument in expression.arguments:
            self.resolve(argument)

    def visit_get_expr(self, expression: nodes.Get):
        self.resolve(expression.object)

    def visit_grouping_expr(self, expression: nodes.Grouping):
        self.resolve(expression.expression)

    def visit_literal_expr(self, expression: nodes.Literal):
        pass

    def visit_logical_expr(self, expression: nodes.Logical):
        self.resolve(expression.left)
        self.resolve(expression.right)

    def visit_set_expr(self, expression: nodes.Set):
        self.resolve(expression.value)
        self.resolve(expression.object)

    def visit_super_expr(self, expression: nodes.Super):
        if self.current_class == ClassType.NONE:
            self.errors.append(
                (
                    expression.keyword,
                    "Can't use 'super' outside of a class.",
                ),
            )
        elif self.current_class != ClassType.SUBCLASS:
            self.errors.append(
                (
                    expression.keyword,
                    "Can't use 'super' in a class with no superclass.",
                ),
            )

        self.resolve_local(expression, expression.keyword)

    def visit_this_expr(self, expression: nodes.This):
        if self.current_class == ClassType.NONE:
            self.errors.append(
                (
                    expression.keyword,
                    "Can't use 'this' outside of a class.",
                ),
            )
        self.resolve_local(expression, expression.keyword)

    def visit_unary_expr(self, expression: nodes.Unary):
        self.resolve(expression.right)

    def visit_variable_expr(self, expression: nodes.Variable):
        if self.scopes and self.current_scope.get(expression.name.lexeme) is False:
            self.errors.append(
                (
                    expression.name,
                    "Can't read local variable in its own initializer.",
                ),
            )
        self.resolve_local(expression, expression.name)



from __future__ import annotations

from enum import Enum
from enum import auto
from typing import Any


class TokenType(Enum):
    # Single-character tokens.
    LEFT_PAREN = auto()
    RIGHT_PAREN = auto()
    LEFT_BRACE = auto()
    RIGHT_BRACE = auto()
    COMMA = auto()
    DOT = auto()
    MINUS = auto()
    PLUS = auto()
    SEMICOLON = auto()
    SLASH = auto()
    STAR = auto()

    # One or two character tokens.
    BANG = auto()
    BANG_EQUAL = auto()
    EQUAL = auto()
    EQUAL_EQUAL = auto()
    LESS = auto()
    LESS_EQUAL = auto()
    GREATER = auto()
    GREATER_EQUAL = auto()

    # Literals.
    IDENTIFIER = auto()
    STRING = auto()
    NUMBER = auto()

    # Keywords.
    AND = auto()
    CLASS = auto()
    ELSE = auto()
    FALSE = auto()
    FUN = auto()
    FOR = auto()
    IF = auto()
    NIL = auto()
    OR = auto()
    PRINT = auto()
    RETURN = auto()
    SUPER = auto()
    THIS = auto()
    TRUE = auto()
    VAR = auto()
    WHILE = auto()

    EOF = auto()


class Token:
    def __init__(self, type: TokenType, lexeme: str, literal: Any, line: int):
        self.type = type
        self.lexeme = lexeme
        self.literal = literal
        self.line = line

    def __str__(self):
        return f"{self.type} {self.lexeme} {self.literal}"

    def __repr__(self):
        return (
            f"Token(type={self.type.name}, "
            f"lexeme={self.lexeme}, "
            f"literal={self.literal}, "
            f"line={self.line})"
        )



from __future__ import annotations

from loxygen import nodes
from loxygen.environment import Environment
from loxygen.exceptions import LoxRunTimeError
from loxygen.runtime import Callable
from loxygen.runtime import Clock
from loxygen.runtime import LoxClass
from loxygen.runtime import LoxFunction
from loxygen.runtime import LoxInstance
from loxygen.runtime import Return
from loxygen.token import Token
from loxygen.token import TokenType


class Interpreter(nodes.Visitor):
    def __init__(self):
        self.globals = Environment()
        self.globals.define("clock", Clock())
        self.locals = {}

        self.env = self.globals

    def visit_literal_expr(self, expression: nodes.Literal):
        return expression.value

    def visit_logical_expr(self, expression: nodes.Logical):
        left = self.evaluate(expression.left)
        if expression.operator.type == TokenType.OR:
            if self.is_truthy(left):
                return left
        else:
            if not self.is_truthy(left):
                return left

        return self.evaluate(expression.right)

    def visit_set_expr(self, expression: nodes.Set):
        object = self.evaluate(expression.object)
        if not isinstance(object, LoxInstance):
            raise LoxRunTimeError(expression.name, "Only instances have fields.")

        value = self.evaluate(expression.value)
        object.set(expression.name, value)

        return value

    def visit_super_expr(self, expression: nodes.Super):
        distance = self.locals.get(expression)
        superclass = self.env.get_at(distance, "super")
        object = self.env.get_at(distance - 1, "this")
        method = superclass.find_method(expression.method.lexeme)
        if method is None:
            raise LoxRunTimeError(
                expression.method,
                f"Undefined property '{expression.method.lexeme}'.",
            )

        return method.bind(object)

    def visit_this_expr(self, expression: nodes.This):
        return self.look_up_variable(expression.keyword, expression)

    def visit_unary_expr(self, expression: nodes.Unary):
        right = self.evaluate(expression.right)
        match expression.operator.type:
            case TokenType.BANG:
                return not self.is_truthy(right)
            case TokenType.MINUS:
                self.check_number_operand(expression.operator, right)
                return -right

    @staticmethod
    def check_number_operand(operator, operand):
        if not isinstance(operand, float):
            raise LoxRunTimeError(operator, "Operand must be a number.")

    def visit_variable_expr(self, expression: nodes.Variable):
        return self.look_up_variable(expression.name, expression)

    def look_up_variable(self, name: Token, expression: nodes.Expr):
        distance = self.locals.get(expression)
        if distance is not None:
            return self.env.get_at(distance, name.lexeme)
        return self.globals.get(name)

    def visit_binary_expr(self, expression: nodes.Binary):
        left = self.evaluate(expression.left)
        right = self.evaluate(expression.right)

        match expression.operator.type:
            case TokenType.GREATER:
                self.check_number_operands(expression.operator, left, right)
                return left > right
            case TokenType.GREATER_EQUAL:
                self.check_number_operands(expression.operator, left, right)
                return left >= right
            case TokenType.LESS:
                self.check_number_operands(expression.operator, left, right)
                return left < right
            case TokenType.LESS_EQUAL:
                self.check_number_operands(expression.operator, left, right)
                return left <= right
            case TokenType.MINUS:
                self.check_number_operands(expression.operator, left, right)
                return left - right
            case TokenType.PLUS:
                if isinstance(left, float) and isinstance(right, float):
                    return left + right
                if isinstance(left, str) and isinstance(right, str):
                    return left + right
                raise LoxRunTimeError(
                    expression.operator,
                    "Operands must be two numbers or two strings.",
                )
            case TokenType.SLASH:
                self.check_number_operands(expression.operator, left, right)
                return left / right if right else float("nan")
            case TokenType.STAR:
                self.check_number_operands(expression.operator, left, right)
                return left * right
            case TokenType.BANG_EQUAL:
                return not self.is_equal(left, right)
            case TokenType.EQUAL_EQUAL:
                return self.is_equal(left, right)

        return None

    @staticmethod
    def check_number_operands(operator, left, right):
        if not isinstance(left, float) or not isinstance(right, float):
            raise LoxRunTimeError(operator, "Operands must be numbers.")

    def visit_call_expr(self, expression: nodes.Call):
        callee = self.evaluate(expression.callee)
        arguments = [self.evaluate(argument) for argument in expression.arguments]

        if not isinstance(callee, Callable):
            raise LoxRunTimeError(
                expression.paren,
                "Can only call functions and classes.",
            )

        if (nb_args := len(arguments)) != (arity := callee.arity()):
            raise LoxRunTimeError(
                expression.paren,
                f"Expected {arity} arguments but got {nb_args}.",
            )

        return callee.call(self, arguments)

    def visit_get_expr(self, expression: nodes.Get):
        object = self.evaluate(expression.object)
        if isinstance(object, LoxInstance):
            return object.get(expression.name)

        raise LoxRunTimeError(expression.name, "Only instances have properties.")

    def visit_grouping_expr(self, expression: nodes.Grouping):
        return self.evaluate(expression.expression)

    @staticmethod
    def is_truthy(obj):
        if obj is None:
            return False
        if isinstance(obj, bool):
            return obj
        return True

    @staticmethod
    def is_equal(obj1, obj2):
        if isinstance(obj1, float) and isinstance(obj2, bool):
            return False
        if isinstance(obj1, bool) and isinstance(obj2, float):
            return False

        return obj1 == obj2

    def evaluate(self, expression: nodes.Expr):
        return expression.accept(self)

    def execute(self, statement: nodes.Stmt):
        statement.accept(self)

    def resolve(self, expression: nodes.Expr, depth: int):
        self.locals[expression] = depth

    def execute_block(self, statements, environment):
        enclosing = self.env
        try:
            self.env = environment
            for statement in statements:
                self.execute(statement)
        finally:
            self.env = enclosing

    def visit_block_stmt(self, stmt: nodes.Block):
        self.execute_block(stmt.statements, Environment(self.env))

    def visit_class_stmt(self, stmt: nodes.Class):
        superclass = None
        if stmt.superclass is not None:
            superclass = self.evaluate(stmt.superclass)
            if not isinstance(superclass, LoxClass):
                raise LoxRunTimeError(
                    stmt.superclass.name,
                    "Superclass must be a class.",
                )

        self.env.define(stmt.name.lexeme, None)
        if stmt.superclass is not None:
            self.env = Environment(self.env)
            self.env.define("super", superclass)

        methods = {}
        for method in stmt.methods:
            is_initializer = method.name.lexeme == "init"
            function = LoxFunction(method, self.env, is_initializer)
            methods[method.name.lexeme] = function
        cls = LoxClass(stmt.name.lexeme, superclass, methods)

        if stmt.superclass is not None:
            self.env = self.env.enclosing

        self.env.assign(stmt.name, cls)

    def visit_expression_stmt(self, stmt: nodes.Expression) -> None:
        self.evaluate(stmt.expression)

    def visit_function_stmt(self, stmt: nodes.Function) -> None:
        function = LoxFunction(stmt, self.env, False)
        self.env.define(stmt.name.lexeme, function)

    def visit_if_stmt(self, stmt: nodes.If) -> None:
        if self.is_truthy(self.evaluate(stmt.condition)):
            self.execute(stmt.then_branch)
        elif stmt.else_branch is not None:
            self.execute(stmt.else_branch)

    def visit_print_stmt(self, stmt: nodes.Print) -> None:
        value = self.evaluate(stmt.expression)
        print(self.stringify(value))

    def visit_return_stmt(self, stmt: nodes.Return) -> None:
        if (value := stmt.value) is not None:
            value = self.evaluate(stmt.value)

        raise Return(value)

    def visit_var_stmt(self, stmt: nodes.Var) -> None:
        value = None
        if stmt.initializer is not None:
            value = self.evaluate(stmt.initializer)
        self.env.define(stmt.name.lexeme, value)

    def visit_while_stmt(self, stmt: nodes.While):
        while self.is_truthy(self.evaluate(stmt.condition)):
            self.execute(stmt.body)

    def visit_assign_expr(self, expression: nodes.Assign):
        value = self.evaluate(expression.value)
        distance = self.locals.get(expression)
        if distance is not None:
            self.env.assign_at(distance, expression.name, value)
        else:
            self.globals.assign(expression.name, value)

        return value

    def interpret(self, *statements: nodes.Stmt):
        for statement in statements:
            self.execute(statement)

    def stringify(self, obj):
        if obj is None:
            return "nil"
        if isinstance(obj, bool):
            return str(obj).lower()
        if isinstance(obj, float) and (repr := str(obj)).endswith(".0"):
            return repr[:-2]

        return str(obj)



from __future__ import annotations

import re
import shlex
import subprocess
import traceback
from collections import Counter
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from pytest import CallInfo
from pytest import Config
from pytest import Dir
from pytest import ExceptionInfo
from pytest import ExitCode
from pytest import TestReport

from contract.contract import LoxStatus

if TYPE_CHECKING:
    from _pytest._code.code import TerminalRepr
    from _pytest.terminal import TerminalReporter


DEFAULT_INTERPRETER = "loxygen"
DEFAULT_SKIP_DIRS = ["benchmark", "scanning", "limit", "expressions"]


@dataclass(frozen=True)
class Option:
    name: str
    help: str
    ini: dict[str, str | list[str]]
    cli: dict[str, str]
    processor: Callable | None


OPTIONS: dict[str, Option] = {
    "interpreter_cmd": Option(
        name="interpreter_cmd",
        help="The command to run the interpreter.",
        ini={"type": "string", "default": DEFAULT_INTERPRETER},
        cli={},
        processor=shlex.split,
    ),
    "skip_dirs": Option(
        name="skip_dirs",
        help="Skips tests located within the specified directory names.",
        ini={"type": "args", "default": DEFAULT_SKIP_DIRS},
        cli={"action": "append"},
        processor=None,
    ),
}


INTERPRETER_CMD_KEY = pytest.StashKey[list[str]]()
SKIP_DIRS_KEY = pytest.StashKey[list[str]]()


@dataclass
class LoxEvent:
    status: LoxStatus
    text: str


@dataclass
class ExpectedLoxEvent(LoxEvent):
    lineno: int


class LoxTestError(Exception):
    """Base exception for all errors raised by the loxtest plugin."""

    pass


class FailedTestException(LoxTestError):
    """Indicates a mismatch between test output and expected output."""

    def __init__(self, failed_lines: tuple[int], *args):
        super().__init__(*args)
        self.failed_lines = failed_lines


class BackEndError(LoxTestError):
    """Indicates a failure while trying to execute the lox interpreter."""

    def __init__(self, error: str, *args):
        super().__init__(*args)
        self.error = error


class TestItem(pytest.Item):
    def __init__(self, name: str, expected, **kwargs):
        super().__init__(name, **kwargs)
        self.name: str = name
        self.expected: list[ExpectedLoxEvent] = expected
        self.output: list[LoxEvent] = []

    def runtest(self):
        self.run_lox()

        if (output_len := len(self.output)) != (expected_len := len(self.expected)):
            raise FailedTestException(
                (-1,),
                f"Mismatch in number of output lines: expected {expected_len},"
                f" but got {output_len}.",
            )

        failed_lines = tuple(
            expected.lineno
            for output, expected in zip(self.output, self.expected)
            if (output.status != expected.status) or (output.text != expected.text)
        )

        if len(failed_lines):
            raise FailedTestException(failed_lines)

    def run_lox(self):
        cmd = self.config.stash[INTERPRETER_CMD_KEY] + [self.path]
        try:
            process = subprocess.run(cmd, capture_output=True, text=True)
        except OSError:
            raise BackEndError(traceback.format_exc()) from None

        if process.stdout:
            self.output = [
                LoxEvent(LoxStatus.OK, line.strip()) for line in process.stdout.splitlines()
            ]

        if process.returncode != 0:
            if process.returncode in LoxStatus:
                self.output.extend(
                    LoxEvent(LoxStatus(process.returncode), line.strip())
                    for line in process.stderr.splitlines()
                )
            else:
                raise BackEndError(process.stderr)

    def repr_failure(self, excinfo: ExceptionInfo, *args, **kwargs) -> str | TerminalRepr:
        if isinstance(excinfo.value, FailedTestException):
            return self.colorize(*excinfo.value.failed_lines)
        if isinstance(excinfo.value, BackEndError):
            return excinfo.value.error

        return super().repr_failure(excinfo, *args, **kwargs)

    def add_result(self) -> list[str]:
        text = self.path.read_text().splitlines()

        for result, output in zip(self.expected, self.output):
            text[result.lineno] = text[result.lineno] + f" // output: {output.text}"

        return text

    def colorize(self, *indexes: int) -> str:
        colors = {"red": 91, "green": 92}
        text = self.add_result()
        if indexes == (-1,):
            indexes = tuple(range(len(text)))
        for result in self.expected:
            color = colors["red"] if result.lineno in indexes else colors["green"]
            text[result.lineno] = f"\033[{color}m{text[result.lineno]}\033[0m"

        return "\n".join(text)

    def reportinfo(self):
        return self.path, 0, self.name


class LoxFile(pytest.File):
    @staticmethod
    def process_match(result: re.Match):
        for group, output in result.groupdict().items():
            if output is not None:
                return group, output

    def parse_test(self) -> list[ExpectedLoxEvent]:
        pattern = re.compile(
            rf"// expect: (?P<{LoxStatus.OK.name.lower()}>.*)|"
            rf"// (?P<{LoxStatus.STATIC_ERROR.name.lower()}>\[line \d+] Error.*)|"
            rf"// expect runtime error: (?P<{LoxStatus.RUNTIME_ERROR.name.lower()}>(.*))",
        )

        results = [
            (lineno, *self.process_match(result))
            for lineno, line in enumerate(self.path.read_text().splitlines())
            if (result := re.search(pattern, line))
        ]

        return [
            ExpectedLoxEvent(LoxStatus[group.upper()], output.strip(), lineno)
            for lineno, group, output in results
        ]

    def collect(self):
        item = TestItem.from_parent(
            self,
            name=self.path.stem,
            expected=None,
        )

        root_path = self.config.rootpath
        parts = self.path.relative_to(root_path).parent.parts
        skip_dirs = self.parent.config.stash[SKIP_DIRS_KEY]

        if not set(parts).isdisjoint(set(skip_dirs)):
            skipped_dir = set(parts).intersection(skip_dirs).pop()
            reason = f"Test located in a skipped directory: {skipped_dir}"
            item.add_marker(pytest.mark.skip(reason=reason))
        else:
            item.expected = self.parse_test()

        yield item


def pytest_addoption(parser: pytest.Parser):
    for option in OPTIONS.values():
        parser.addini(option.name, option.help, **option.ini)
        parser.addoption(f"--{option.name}", help=option.help, **option.cli)


def get_value(config: Config, option: Option):
    name = option.name
    if (value := config.getoption(name)) is None:
        value = config.getini(name)
    if (processor := option.processor) is not None:
        value = processor(value)

    return value


def pytest_configure(config: Config):
    option = OPTIONS["interpreter_cmd"]
    value = get_value(config, option)
    config.stash[INTERPRETER_CMD_KEY] = value

    option = OPTIONS["skip_dirs"]
    value = get_value(config, option)
    config.stash[SKIP_DIRS_KEY] = value


def pytest_collect_file(parent: Dir, file_path: Path):
    if file_path.suffix == ".lox":
        return LoxFile.from_parent(
            parent,
            path=file_path,
        )

    return None


def pytest_runtest_makereport(item: TestItem, call: CallInfo) -> TestReport:
    report = TestReport.from_item_and_call(item, call)
    if call.when == "call" and call.excinfo:
        if isinstance(call.excinfo.value, BackEndError):
            report.user_properties.append(
                ("BackEndError", call.excinfo.value.error.splitlines()[-1])
            )

    return report


def pytest_terminal_summary(
    terminalreporter: TerminalReporter,
    exitstatus: ExitCode,
    config: Config,
):
    if config.getoption("--collect-only", default=False):
        return None

    failed_reports: list[TestReport] = terminalreporter.getreports("failed")
    if not (failed_reports := [report for report in failed_reports if report.when == "call"]):
        return None

    backend_errors = [
        message
        for report in failed_reports
        for (error, message) in report.user_properties
        if error == "BackEndError"
    ]

    if backend_errors:
        terminalreporter.ensure_newline()
        terminalreporter.section("python backend errors summary", sep="-", blue=True, bold=True)
        for error, count in Counter(backend_errors).items():
            terminalreporter.line(f"{error} (occurred {count} time{'s' if count > 1 else ''})")



from __future__ import annotations

from loxygen.exceptions import LoxRunTimeError
from loxygen.token import Token


class Environment:
    def __init__(self, enclosing=None):
        self.values = {}
        self.enclosing = enclosing

    def define(self, name: str, value: object) -> None:
        self.values[name] = value

    def ancestor(self, distance):
        environment = self
        for _ in range(distance):
            environment = environment.enclosing
        return environment

    def get_at(self, distance: int, name: str):
        return self.ancestor(distance).values[name]

    def assign_at(self, distance: int, name: Token, value):
        self.ancestor(distance).values[name.lexeme] = value

    def get(self, name: Token) -> object:
        try:
            return self.values[name.lexeme]
        except KeyError:
            pass

        if self.enclosing is not None:
            return self.enclosing.get(name)

        raise LoxRunTimeError(name, f"Undefined variable '{name.lexeme}'.")

    def assign(self, name: Token, value: object):
        if name.lexeme in self.values:
            self.values[name.lexeme] = value
            return

        if self.enclosing is not None:
            self.enclosing.assign(name, value)
            return

        raise LoxRunTimeError(name, f"Undefined variable '{name.lexeme}'.")
